# -*- coding: utf-8 -*-
from marshmallow import Schema, fields

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"


# module wide constant utilized in JSON serialization to specify key for JSON schema info, if required
BGEdge_JSON_SCHEMA_JSON_KEY = "_py__bg_edge_json_schema"


class BGEdge(object):
    """ A wrapper class for edges in :class:`bg.breakpoint_graph.BreakpointGraph`

    Is not stored on its own in the :class:`bg.breakpoint_graph.BreakpointGraph`, but is rather can be supplied to work with and is returned if retrieval is performed.
    BGEdge represents an undirected edge, thus distinction between :attr:`BGEdge.vertex1` and :attr:`BGEdge.vertex2` attributes is just from identities perspective, not from the order perspective.

    This class supports th following attributes, that cary information about multi-color for this edge, as well as vertices, its is attached to:

    *   :attr:`BGEdge.vertex1`: a first vertex to be utilized in :class:`bg.breakpoint_graph.BreakpointGraph`
    *   :attr:`BGEdge.vertex2`: a second vertex to be utilized in :class:`bg.breakpoint_graph.BreakpointGraph`

    Main operations:

    *   ``==``
    *   :meth:`BGEdge.merge`: produces a new BGEdge with multi-color information being merged from them
    """

    class BGEdgeJSONSchema(Schema):
        """ Marshmallow powered JSON schema used for serialization / deserialization of edge object """
        _py__bg_edge_json_schema = fields.String(attribute="json_schema_name")
        vertex1_id = fields.Int(attribute="vertex1_json_id", required=True)
        vertex2_id = fields.Int(attribute="vertex2_json_id", required=True)
        multicolor = fields.List(fields.Int, attribute="colors_json_ids", allow_none=False, required=True)

        def make_object(self, data):
            if "vertex1_json_id" not in data:
                raise ValueError("Error during edge serialization. \"vertex1_id\" key is not present in json object")
            vertex1 = data["vertex1_json_id"]
            if "vertex2_json_id" not in data:
                raise ValueError("Error during edge serialization. \"vertex2_id\" key is not present in json object")
            vertex2 = data["vertex2_json_id"]
            if "colors_json_ids" not in data:
                raise ValueError("Error during edge serialization. \"multicolor\" key is not present in json object")
            multicolor = data["colors_json_ids"]
            return BGEdge(vertex1=vertex1, vertex2=vertex2, multicolor=multicolor)

    # class wide variable for json serialization/deserialization. Created once for a whole class, as thousands of objects
    # undergo serialization / deserialization, and schema instantiation in each case would require additional resources
    json_schema = BGEdgeJSONSchema()

    def __init__(self, vertex1, vertex2, multicolor):
        """ Initialization of :class:`BGEdge` object.

        :param vertex1: vertex the edges is outgoing from
        :type vertex1: any hashable python object
        :param vertex2: vertex the edges is ingoing to
        :type vertex2: any hashable python object
        :param multicolor: multicolor that this single edge shall posses
        :type multicolor: :class:`bg.multicolor.Multicolor`
        :return: a new instance of :class:`BGEdge`
        :rtype: :class:`BGEdge`
        """
        self.vertex1 = vertex1
        self.vertex2 = vertex2
        self.multicolor = multicolor

    @classmethod
    def merge(cls, edge1, edge2):
        """ Merges multi-color information from two supplied :class:`BGEdge` instances into a new :class:`BGEdge`

        Since :class:`BGEdge` represents an undirected edge, created edge's vertices are assign according to the order in first supplied edge.

        Accounts for subclassing.

        :param edge1: first out of two edge information from which is to be merged into a new one
        :param edge2: second out of two edge information from which is to be merged into a new one
        :return: a new undirected with multi-color information merged from two supplied :class:`BGEdge` objects
        :raises: ``ValueError``
        """
        if edge1.vertex1 != edge2.vertex1 and edge1.vertex1 != edge2.vertex2:
            raise ValueError("Edges to be merged do not connect same vertices")
        forward = edge1.vertex1 == edge2.vertex1
        if forward and edge1.vertex2 != edge2.vertex2:
            raise ValueError("Edges to be merged do not connect same vertices")
        elif not forward and edge1.vertex2 != edge2.vertex1:
            raise ValueError("Edges to be merged do not connect same vertices")
        return cls(vertex1=edge1.vertex1, vertex2=edge1.vertex2, multicolor=edge1.multicolor + edge2.multicolor)

    def __eq__(self, other):
        """ Implementation of ``==`` operation for :class:`BGEdge`

        Checks if current :class:`BGEdge` instance comply in terms of vertices set with the supplied :class:`BGEdge`, and then checks the equality of :attr:`BGEdge.multicolor` attributes in respective objects.
         :class:`BGEdge` does not equal to non-:class:`BGEdge` objects

        :param other: object to compare current :class:`BGEdge` to
        :type other: any python object
        :return: flag of equality if current :class:`BGEdge` object equals to the supplied one
        :rtype: ``Boolean``
        """
        if not isinstance(other, BGEdge):
            return False
        if self.vertex1 != other.vertex1 and self.vertex1 != other.vertex2:
            return False
        multicolor_equality = self.multicolor == other.multicolor
        if self.vertex1 == other.vertex1:
            return self.vertex2 == other.vertex2 and multicolor_equality
        else:
            return self.vertex2 == other.vertex1 and multicolor_equality

    def __getattr__(self, item):
        if item.startswith("is_") and item.endswith("_edge"):
            lookup = item[3:-5]
            vertex_lookup = "is_" + lookup + "_vertex"
            if getattr(self.vertex1, vertex_lookup) or getattr(self.vertex2, vertex_lookup):
                return True
            return False
        return super().__getattribute__(item)

    @property
    def json_schema_name(self):
        """ When genome is serialized information about JSON schema of such serialization can be recorded,
        and this property provides access to it
        """
        return self.json_schema.__class__.__name__

    @staticmethod
    def __vertex_json_id(vertex):
        """ A proxy property based access to vertices in current edge

        When edge is serialized to JSON object, no explicit object for its vertices are created, but rather they are referenced
        by special vertex json_ids.
        """
        if hasattr(vertex, "json_id"):
            return vertex.json_id
        return hash(vertex)

    @property
    def vertex1_json_id(self):
        """ First vertex json id access """
        return self.__vertex_json_id(self.vertex1)

    @property
    def vertex2_json_id(self):
        """ Second vertex json id access """
        return self.__vertex_json_id(self.vertex2)

    @property
    def colors_json_ids(self):
        """ A proxy property based access to vertices in current edge

        When edge is serialized to JSON object, no explicit object for its multicolor is created, but rather all colors,
        taking into account their multiplicity, are referenced by their json_ids.
        """
        return [color.json_id if hasattr(color, "json_id") else hash(color) for color in self.multicolor.multicolors.elements()]

    def to_json(self, schema_info=True):
        """ JSON serialization method that accounts for a possibility of field filtration and schema specification """
        old_exclude_fields = self.json_schema.exclude
        new_exclude_fields = list(old_exclude_fields)
        if not schema_info:
            new_exclude_fields.append(BGEdge_JSON_SCHEMA_JSON_KEY)
        # monkey patch schema `exclude` attribute to ignore some fields in result json object
        self.json_schema.exclude = new_exclude_fields
        result = self.json_schema.dump(self).data
        # reverse the result of monkey patching
        self.json_schema.exclude = old_exclude_fields
        return result

    @classmethod
    def from_json(cls, data, json_schema_class=None):
        """ JSON deserialization method that retrieves edge instance from its json representation

        If specific json schema is provided, it is utilized, and if not, a class specific is used
        """
        schema = cls.json_schema if json_schema_class is None else json_schema_class()
        return schema.load(data).data



import io
import os
from collections import Counter

from bg import BreakpointGraph, Multicolor
from bg.bg_io import GRIMMReader, GRIMMWriter
from bg.genome import BGGenome
from bg.vertices import TaggedBlockVertex, TaggedInfinityVertex

__author__ = 'Sergey Aganezov'
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

import unittest


class GRIMMReaderTestCase(unittest.TestCase):
    def test_is_genome_declaration_string(self):
        # string is named as genome declaration string if its first non empty element is ">"
        # genome name has to be specified after the ">" char, empty genome name is forbidden
        self.assertTrue(GRIMMReader.is_genome_declaration_string(">genome"))
        self.assertTrue(GRIMMReader.is_genome_declaration_string("    >genome"))
        self.assertTrue(GRIMMReader.is_genome_declaration_string("  \t  >genome"))
        self.assertTrue(GRIMMReader.is_genome_declaration_string(">genome   \t"))
        self.assertTrue(GRIMMReader.is_genome_declaration_string(">genome   "))
        self.assertTrue(GRIMMReader.is_genome_declaration_string("   >genome   "))
        self.assertFalse(GRIMMReader.is_genome_declaration_string("\tt   >genome"))
        self.assertFalse(GRIMMReader.is_genome_declaration_string("  t\t>genome"))
        self.assertFalse(GRIMMReader.is_genome_declaration_string("  t>genome"))
        self.assertFalse(GRIMMReader.is_genome_declaration_string("genome"))
        self.assertFalse(GRIMMReader.is_genome_declaration_string(">"))
        self.assertFalse(GRIMMReader.is_genome_declaration_string("     >   "))
        self.assertFalse(GRIMMReader.is_genome_declaration_string("     >"))
        self.assertFalse(GRIMMReader.is_genome_declaration_string(">   "))

    def test_parse_genome_declaration_string(self):
        # genome declaration string is parsed, by stripping the string from the right
        # and retrieving the string after the ">" character
        self.assertEqual(GRIMMReader.parse_genome_declaration_string(">genome"), BGGenome("genome"))
        self.assertEqual(GRIMMReader.parse_genome_declaration_string("  >genome  "), BGGenome("genome"))
        self.assertEqual(GRIMMReader.parse_genome_declaration_string(">genome__genome"), BGGenome("genome__genome"))
        self.assertEqual(GRIMMReader.parse_genome_declaration_string(">genome>genome"), BGGenome("genome>genome"))
        self.assertEqual(GRIMMReader.parse_genome_declaration_string(">genome.!/.#4"), BGGenome("genome.!/.#4"))

    def test_parse_data_string_error(self):
        # data string must contain a fragment termination symbol ($ or @)
        # and must contain space separated gene order information before fragment termination symbol
        data_string_1 = "   a b c d e    "
        data_string_2 = ""
        data_string_3 = " a -b -c d -e "
        data_string_4 = "$"
        data_string_5 = "@"
        data_string_6 = "@ a d s d"
        data_string_7 = "$a d s d"
        data_string_8 = "$-a d s d"
        data_string_9 = "@+a d s d"
        data_string_10 = "a b - -c d e $"
        for data_string in [data_string_1, data_string_2, data_string_3, data_string_4, data_string_5,
                            data_string_6, data_string_7, data_string_8, data_string_9, data_string_10]:
            with self.assertRaises(ValueError):
                GRIMMReader.parse_data_string(data_string)

    def test_parse_data_string_correct(self):
        # data string is parsed by getting information about genes order and individual orientations for each block (gene)
        # string based processing if performed
        # if no orientation is specified explicitly, positive orientation is assumed
        data_string = "a $"
        result = GRIMMReader.parse_data_string(data_string)
        self.assertEqual(result[0], "$")
        self.assertEqual(result[1][0][0], "+")
        self.assertEqual(result[1][0][1], "a")
        self.assertEqual(len(result[0]), 1)
        self.assertEqual(len(result[1]), 1)

        data_string = "a @"
        result = GRIMMReader.parse_data_string(data_string)
        self.assertEqual(result[0], "@")
        self.assertEqual(result[1][0][0], "+")
        self.assertEqual(result[1][0][1], "a")
        self.assertEqual(len(result[0]), 1)
        self.assertEqual(len(result[1]), 1)

        data_string = "     a -b c -d @ e f     "
        result = GRIMMReader.parse_data_string(data_string)
        self.assertEqual(result[0], "@")
        reference_genes = ["a", "b", "c", "d"]
        result_genes = [gene[1] for gene in result[1]]
        reference_signs = ["+", "-", "+", "-"]
        result_signs = [gene[0] for gene in result[1]]
        self.assertListEqual(result_genes, reference_genes)
        self.assertListEqual(result_signs, reference_signs)

        data_string = "     a -b +c -d $ e f     "
        result = GRIMMReader.parse_data_string(data_string)
        self.assertEqual(result[0], "$")
        reference_genes = ["a", "b", "c", "d"]
        result_genes = [gene[1] for gene in result[1]]
        reference_signs = ["+", "-", "+", "-"]
        result_signs = [gene[0] for gene in result[1]]
        self.assertListEqual(result_genes, reference_genes)
        self.assertListEqual(result_signs, reference_signs)

        data_string = "     a -b c -d @ e f $ g -h    "
        result = GRIMMReader.parse_data_string(data_string)
        self.assertEqual(result[0], "@")
        reference_genes = ["a", "b", "c", "d"]
        result_genes = [gene[1] for gene in result[1]]
        reference_signs = ["+", "-", "+", "-"]
        result_signs = [gene[0] for gene in result[1]]
        self.assertListEqual(result_genes, reference_genes)
        self.assertListEqual(result_signs, reference_signs)

    def test_get_list_of_edges_no_repeat_blocks(self):
        # depending on the fragment type adjacencies to be considered in BreakpointGraph are differ
        # in case of circular genome, additional adjacency is added between to outermost vertices
        # in case of linear genome, two extremity (infinity) vertices are appended to the start and end of the vertices list
        parsed_data = ("@", [("+", "a"), ("-", "b"), ("-", "a")])
        result = GRIMMReader.get_edges_from_parsed_data(parsed_data)
        reference = [(TaggedBlockVertex("at"), TaggedBlockVertex("at")),
                     (TaggedBlockVertex("ah"), TaggedBlockVertex("bh")),
                     (TaggedBlockVertex("bt"), TaggedBlockVertex("ah"))]
        self.assertDictEqual(Counter(result), Counter(reference))

        parsed_data = ("$", [("+", "a"), ("-", "b"), ("-", "a")])
        result = GRIMMReader.get_edges_from_parsed_data(parsed_data)
        reference = [(TaggedInfinityVertex("at"), TaggedBlockVertex("at")),
                     (TaggedBlockVertex("ah"), TaggedBlockVertex("bh")),
                     (TaggedBlockVertex("bt"), TaggedBlockVertex("ah")),
                     (TaggedBlockVertex("at"), TaggedInfinityVertex("at"))]
        self.assertDictEqual(Counter(result), Counter(reference))

        parsed_data = ("@", [("+", "a")])
        result = GRIMMReader.get_edges_from_parsed_data(parsed_data)
        reference = [(TaggedBlockVertex("ah"), TaggedBlockVertex("at"))]
        self.assertDictEqual(Counter(result), Counter(reference))

        parsed_data = ("$", [("-", "a"), ("-", "a")])
        result = GRIMMReader.get_edges_from_parsed_data(parsed_data)
        reference = [(TaggedInfinityVertex("ah"), TaggedBlockVertex("ah")),
                     (TaggedBlockVertex("at"), TaggedBlockVertex("ah")),
                     (TaggedBlockVertex("at"), TaggedInfinityVertex("at"))]
        self.assertDictEqual(Counter(result), Counter(reference))

    def test_get_list_of_edges_repeat_blocks_at_extremities(self):
        # if a fragment starts and / or ends with repeat block, that is denoted in a form of
        # repeat_block_name__repeat
        # IF this block is preserved as a block (i.e. located inside the fragment),
        # it will be transformed into a block "repeat_block_name" with
        #   an empty tag -- value "repeat" -- None
        # IF the block is not preserved (flanking repeat), it will be dismissed and its outermost extremity
        #   will be used to make its name (i.e. "repeat_block_name"h and it will be added as a tag to the infinity vertex
        #   created for a linear fragment
        # such information shall be recorded in respective infinity vertex, if a fragment is linear,
        # and in a normal tagged vertex, if fragment is circular

        # single repeat on the left extremity of linear fragment
        parsed_data = ("$", [("+", "a__repeat"), ("-", "b"), ("+", "c__tag:1")])
        result = GRIMMReader.get_edges_from_parsed_data(parsed_data)
        left_iv = TaggedInfinityVertex("bh")
        left_iv.add_tag("repeat", "ah")
        ch_vertex, ct_vertex = TaggedBlockVertex("ch"), TaggedBlockVertex("ct")
        ch_vertex.add_tag("tag", "1")
        ct_vertex.add_tag("tag", "1")
        reference = [(left_iv, TaggedBlockVertex("bh")),
                     (TaggedBlockVertex("bt"), ct_vertex),
                     (ch_vertex, TaggedInfinityVertex("ch"))]
        self.assertDictEqual(Counter(result), Counter(reference))

        # both extremities are "flanked" by repeats
        parsed_data = ("$", [("+", "a__repeat"), ("-", "b"), ("+", "c__tag:1:2"), ("-", "a__repeat")])
        result = GRIMMReader.get_edges_from_parsed_data(parsed_data)
        left_iv = TaggedInfinityVertex("bh")
        left_iv.add_tag("repeat", "ah")
        right_iv = TaggedInfinityVertex("ch")
        right_iv.add_tag("repeat", "ah")
        ch_vertex, ct_vertex = TaggedBlockVertex("ch"), TaggedBlockVertex("ct")
        ch_vertex.add_tag("tag", "1:2")
        ct_vertex.add_tag("tag", "1:2")
        reference = [(left_iv, TaggedBlockVertex("bh")),
                     (TaggedBlockVertex("bt"), ct_vertex),
                     (ch_vertex, right_iv)]
        self.assertDictEqual(Counter(result), Counter(reference))

        # fragment is specified as circular, all repeats shall be treated as normal blocks with half empty tags
        parsed_data = ("@", [("+", "a__repeat"), ("-", "b"), ("+", "c__tag:1"), ("-", "a__repeat")])
        result = GRIMMReader.get_edges_from_parsed_data(parsed_data)
        ch_vertex, ct_vertex = TaggedBlockVertex("ch"), TaggedBlockVertex("ct")
        ct_vertex.add_tag("tag", "1")
        ch_vertex.add_tag("tag", "1")
        ah_vertex, at_vertex = TaggedBlockVertex("ah"), TaggedBlockVertex("at")
        ah_vertex.add_tag("repeat", None)
        at_vertex.add_tag("repeat", None)
        reference = [(ah_vertex, TaggedBlockVertex("bh")),
                     (TaggedBlockVertex("bt"), ct_vertex),
                     (ch_vertex, ah_vertex),
                     (at_vertex, at_vertex)]
        self.assertDictEqual(Counter(result), Counter(reference))

    def test_is_comment_string(self):
        # a sting is considered a comment if it non empty first char is "#"
        self.assertTrue(GRIMMReader.is_comment_string("#"))
        self.assertTrue(GRIMMReader.is_comment_string("     #"))
        self.assertTrue(GRIMMReader.is_comment_string("#    "))
        self.assertTrue(GRIMMReader.is_comment_string("     #    "))
        self.assertTrue(GRIMMReader.is_comment_string("#  aaa  "))
        self.assertFalse(GRIMMReader.is_comment_string("a# "))
        self.assertTrue(GRIMMReader.is_comment_string("    ##  "))

    def test_get_breakpoint_from_file(self):
        # full workflow testing with dummy data
        # correct cases are assumed with all kind of crazy indentation and rubbish data mixed in, but still correct
        data = ["",
                "\t",
                "#comment1",
                ">genome_name_1",
                "      #comment1",
                "a b $",
                "\tc -a @\t",
                "   #comment1   ",
                "\t>genome_name_2",
                "a $",
                "",
                "\n\t"]
        file_like = io.StringIO("\n".join(data))
        result_bg = GRIMMReader.get_breakpoint_graph(file_like)
        self.assertTrue(isinstance(result_bg, BreakpointGraph))
        self.assertEqual(len(list(result_bg.connected_components_subgraphs())), 3)
        self.assertEqual(len(list(result_bg.edges())), 6)
        self.assertEqual(len(list(result_bg.nodes())), 9)
        multicolors = [Multicolor(BGGenome("genome_name_1"), BGGenome("genome_name_2")),
                       Multicolor(BGGenome("genome_name_1")),
                       Multicolor(BGGenome("genome_name_2"))]
        for bgedge in result_bg.edges():
            self.assertTrue(bgedge.multicolor in multicolors)
        infinity_edges = [bgedge for bgedge in result_bg.edges() if bgedge.is_infinity_edge]
        self.assertEqual(len(infinity_edges), 3)

        data = [">genome_1",
                "a $",
                ">genome_2",
                "a b $",
                "# this is a bad genome",
                ">genome_3",
                "a b c $",
                ">genome_4",
                "   # chromosome 1",
                "b c $",
                ">genome_5",
                "c $"]
        file_like = io.StringIO("\n".join(data))
        result_bg = GRIMMReader.get_breakpoint_graph(file_like)
        self.assertTrue(isinstance(result_bg, BreakpointGraph))
        self.assertEqual(len(list(result_bg.connected_components_subgraphs())), 4)
        self.assertEqual(len(list(result_bg.edges())), 8)
        self.assertEqual(len(list(result_bg.nodes())), 12)
        genome1, genome2, genome3 = BGGenome("genome_1"), BGGenome("genome_2"), BGGenome("genome_3")
        genome4, genome5 = BGGenome("genome_4"), BGGenome("genome_5")
        multicolors = [Multicolor(genome1, genome2, genome3),
                       Multicolor(genome1),
                       Multicolor(genome2, genome3),
                       Multicolor(genome2),
                       Multicolor(genome3, genome4),
                       Multicolor(genome3, genome4, genome5),
                       Multicolor(genome4),
                       Multicolor(genome5)]
        for bgedge in result_bg.edges():
            self.assertTrue(bgedge.multicolor in multicolors)
        infinity_edges = [bgedge for bgedge in result_bg.edges() if bgedge.is_infinity_edge]
        self.assertEqual(len(infinity_edges), 6)
        infinity_multicolors = [multicolor for multicolor in multicolors if len(multicolor.multicolors) != 2]
        for bgedge in infinity_edges:
            self.assertTrue(bgedge.multicolor in infinity_multicolors)


class GRIMMWriterTestCase(unittest.TestCase):
    def setUp(self):
        self.genome1 = BGGenome("red")
        self.genome2 = BGGenome("green")
        self.genome3 = BGGenome("blue")
        self.single_genome_bg = BreakpointGraph()
        self.two_genome_bg = BreakpointGraph()
        self.four_genome_bg = BreakpointGraph()

    def _populate_bg(self, data):
        file_like = io.StringIO("\n".join(data))
        bg = GRIMMReader.get_breakpoint_graph(file_like)
        return bg

    def _get_mouse_data(self):
        return [
            ">Mouse",
            "1 2 3 4 $",
            "5 6 7 8 $"
        ]

    def _get_human_data(self):
        return [
            ">Human",
            "1 4 3 2 $",
            "5 -7 -6 8 $"
        ]

    def _get_rat_data(self):
        return [
            ">Rat",
            "0 1 4 5 $",
            "10 12 8 7 $"
        ]

    def _get_chimp_data(self):
        return [
            ">Chimpanzee",
            "5 6 7 8 $",
            "1 -4 -3 -2 $"
        ]

    def _populate_single_genome_bg(self):
        data = self._get_mouse_data()
        bg = self._populate_bg(data=data)
        self.single_genome_bg = bg

    def _populate_two_genomes_bg(self):
        data = self._get_human_data() + self._get_mouse_data()
        bg = self._populate_bg(data=data)
        self.two_genome_bg = bg

    def _populate_four_genomes_bg(self):
        data = self._get_human_data() + self._get_mouse_data() + self._get_chimp_data() + self._get_rat_data()
        bg = self._populate_bg(data=data)
        self.four_genome_bg = bg

    def test_get_grimm_strings_from_breakpoint_graph_single_genome(self):
        self._populate_single_genome_bg()
        grimm_strings = GRIMMWriter.get_grimm_from_breakpoint_graph(bg=self.single_genome_bg)
        self.assertEqual(len(grimm_strings), 3)
        self.assertIn(">Mouse", grimm_strings)
        possibilities_1 = ["1 2 3 4 $", "-4 -3 -2 -1 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_1)))
        possibilities_2 = ["5 6 7 8 $", "-8 -7 -6 -5 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_2)))

    def test_get_grimm_strings_from_breakpoints_graph_two_genomes(self):
        self._populate_two_genomes_bg()
        grimm_strings = GRIMMWriter.get_grimm_from_breakpoint_graph(bg=self.two_genome_bg)
        self.assertEqual(len(grimm_strings), 6)
        self.assertIn(">Mouse", grimm_strings)
        self.assertIn(">Human", grimm_strings)

        possibilities_1 = ["1 2 3 4 $", "-4 -3 -2 -1 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_1)))
        possibilities_2 = ["5 6 7 8 $", "-8 -7 -6 -5 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_2)))

        possibilities_3 = ["1 4 3 2 $", "-2 -3 -4 -1 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_3)))
        possibilities_4 = ["5 -7 -6 8 $", "-8 6 7 -5 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_4)))

    def test_get_grimm_from_breakpoint_graph_four_genomes(self):
        self._populate_four_genomes_bg()
        grimm_strings = GRIMMWriter.get_grimm_from_breakpoint_graph(bg=self.four_genome_bg)

        self.assertEqual(len(grimm_strings), 12)
        self.assertIn(">Mouse", grimm_strings)
        self.assertIn(">Human", grimm_strings)
        self.assertIn(">Rat", grimm_strings)
        self.assertIn(">Chimpanzee", grimm_strings)

        possibilities_1 = ["1 2 3 4 $", "-4 -3 -2 -1 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_1)))
        possibilities_2 = ["5 6 7 8 $", "-8 -7 -6 -5 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_2)))

        possibilities_3 = ["1 4 3 2 $", "-2 -3 -4 -1 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_3)))
        possibilities_4 = ["5 -7 -6 8 $", "-8 6 7 -5 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_4)))

        possibilities_5 = ["0 1 4 5 $", "-5 -4 -1 -0 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_5)))
        possibilities_6 = ["10 12 8 7 $", "-7 -8 -12 -10 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_6)))

        possibilities_7 = ["5 6 7 8 $", "-8 -7 -6 -5 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_7)))
        possibilities_8 = ["1 -4 -3 -2 $", "2 3 4 -1 $"]
        self.assertTrue(any(map(lambda entry: entry in grimm_strings, possibilities_8)))

    def test_output_genomes_as_grimm(self):
        self._populate_four_genomes_bg()
        file_name = "file_name.txt"
        GRIMMWriter.print_genomes_as_grimm_blocks_orders(bg=self.four_genome_bg,
                                                         file_name=file_name)
        try:
            with open(file_name, "rt") as source:
                new_bg = GRIMMReader.get_breakpoint_graph(stream=source,
                                                          merge_edges=True)
                self.assertEqual(len(list(new_bg.nodes())), len(list(self.four_genome_bg.nodes())))
                self.assertEqual(len(list(new_bg.edges())), len(list(self.four_genome_bg.edges())))

                self.assertSetEqual(set(new_bg.nodes()), set(self.four_genome_bg.nodes()))
                self.assertSetEqual(new_bg.get_overall_set_of_colors(),
                                    self.four_genome_bg.get_overall_set_of_colors())

        finally:
            if os.path.exists(file_name):
                os.remove(file_name)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from unittest.mock import Mock
from bg.edge import BGEdge
from bg.multicolor import Multicolor
from bg.kbreak import KBreak
from bg.vertices import BlockVertex, InfinityVertex

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

import unittest


class KBreakTestCase(unittest.TestCase):
    def setUp(self):
        self.v1 = BlockVertex("v1")
        self.v2 = BlockVertex("v2")
        self.v3 = BlockVertex("v3")
        self.v4 = BlockVertex("v4")
        self.inf_v1 = InfinityVertex("v1")
        self.inf_v2 = InfinityVertex("v2")
        self.inf_v3 = InfinityVertex("v3")
        self.inf_v4 = InfinityVertex("v4")

    def test_initialization_incorrect_argument_number(self):
        # required
        # initial set of edges in terms of pairs of vertices
        # desired set of edges in terms of pairs of vertices
        # a multicolor to perform such operation on
        mock_vertex = Mock(spec=BlockVertex)
        mock_multicolor = Mock(spec=Multicolor)
        with self.assertRaises(ValueError):
            KBreak(start_edges=[(mock_vertex, ), (mock_vertex, mock_vertex)],
                   result_edges=[(mock_vertex, mock_vertex), (mock_vertex, mock_vertex)],
                   multicolor=mock_multicolor)
        with self.assertRaises(ValueError):
            KBreak(start_edges=[(mock_vertex, mock_vertex), (mock_vertex, mock_vertex)],
                   result_edges=[(mock_vertex, ), (mock_vertex, mock_vertex)],
                   multicolor=mock_multicolor)
        with self.assertRaises(ValueError):
            KBreak(start_edges=[(mock_vertex, mock_vertex), (mock_vertex, )],
                   result_edges=[(mock_vertex, ), (mock_vertex, mock_vertex)],
                   multicolor=mock_multicolor)

    def test_valid_kbreak_matching(self):
        # checks that second, out of two supplied sets of edges (in terms of vertices pairs)
        # corresponds to a matching on the same set of vertices
        # and degrees of each vertex have to be preserved
        v1, v2, v3, v4 = self.v1, self.v2, self.v3, self.v4
        # regular 2-break
        start_edges = [(v1, v2), (v3, v4)]
        end_edges = [(v1, v3), (v2, v4)]
        self.assertTrue(KBreak.valid_kbreak_matchings(start_edges, end_edges))
        end_edges = [(v1, v3), (v4, v2)]
        # regular 2-break
        self.assertTrue(KBreak.valid_kbreak_matchings(start_edges, end_edges))
        self.assertTrue(KBreak.valid_kbreak_matchings(start_edges, start_edges))
        self.assertTrue(KBreak.valid_kbreak_matchings(end_edges, end_edges))
        # degrees are not consistent
        end_edges = [(v1, v3), (v2, v2)]
        self.assertFalse(KBreak.valid_kbreak_matchings(start_edges, end_edges))
        # degrees are not consistent
        end_edges = [(v1, v3), (v2, v4), (v1, v3)]
        self.assertFalse(KBreak.valid_kbreak_matchings(start_edges, end_edges))
        # degrees are not consistent
        start_edges = [(v1, v2), (v2, v3), (v1, v3)]
        end_edges = [(v1, v2), (v2, v3), (v2, v3)]
        self.assertFalse(KBreak.valid_kbreak_matchings(start_edges, end_edges))
        # valid two break on K3 graph
        end_edges = [(v1, v1), (v2, v2), (v3, v3)]
        self.assertTrue(KBreak.valid_kbreak_matchings(start_edges, end_edges))

    def test_initialization_incorrect_bad_edge_sets(self):
        # case when initialization is overall correct, in terms of cardinality of supplied sets of vertices
        # all arguments are present
        # but supplied edges in terms of vertices do not correspond to correct k-break
        v1, v2, v3, v4 = self.v1, self.v2, self.v3, self.v4
        mock_multicolor = Mock(spec=Multicolor)
        start_edges = [(v1, v2), (v3, v4)]
        end_edges = [(v1, v3), (v2, v2)]
        with self.assertRaises(ValueError):
            KBreak(start_edges=start_edges,
                   result_edges=end_edges,
                   multicolor=mock_multicolor)
        end_edges = [(v1, v3), (v2, v4), (v1, v3)]
        with self.assertRaises(ValueError):
            KBreak(start_edges=start_edges,
                   result_edges=end_edges,
                   multicolor=mock_multicolor)
        start_edges = [(v1, v2), (v2, v3), (v1, v3)]
        end_edges = [(v1, v2), (v2, v3), (v2, v3)]
        with self.assertRaises(ValueError):
            KBreak(start_edges=start_edges,
                   result_edges=end_edges,
                   multicolor=mock_multicolor)

    def test_initialization(self):
        # a correct case of a kbreak initialization, where all correctness checks are successfully passed
        v1, v2, v3, v4 = self.v1, self.v2, self.v3, self.v4
        mock_multicolor = Mock(spec=Multicolor)
        start_edges = [(v1, v2), (v3, v4)]
        end_edges = [(v1, v3), (v2, v4)]
        kbreak = KBreak(start_edges=start_edges,
                        result_edges=end_edges,
                        multicolor=mock_multicolor)
        self.assertListEqual(kbreak.start_edges, start_edges)
        self.assertListEqual(kbreak.result_edges, end_edges)
        self.assertEqual(kbreak.multicolor, mock_multicolor)

if __name__ == '__main__':
    unittest.main()



__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

import unittest
from bg.genome import BGGenome, BGGenome_JSON_SCHEMA_JSON_KEY


class BGGenomeTestCase(unittest.TestCase):
    def test_initialization_incorrect(self):
        # empty genomes are not allowed, a name for genome is mandatory
        with self.assertRaises(TypeError):
            g = BGGenome()

    def test_initialization(self):
        # simple correct initialization
        g = BGGenome("name")
        self.assertEqual(g.name, "name")

    def test_hash(self):
        # hash of genome instance is proxies to hash value of its name
        g = BGGenome("name")
        self.assertEqual(hash(g), hash("name"))

    def test_json_id(self):
        # json id for genome is utilized when genome is serialized to json format and equals to hash value of genome instance
        g = BGGenome("name")
        json_id = g.json_id
        self.assertEqual(json_id, hash(g.name))
        self.assertTrue(isinstance(json_id, int))
        g.name = "name1"
        new_json_id = g.json_id
        self.assertEqual(new_json_id, hash(g.name))
        self.assertTrue(isinstance(json_id, int))
        self.assertNotEqual(json_id, new_json_id)

    def test__eq__(self):
        # two genome are called equal if they are both os same class and their hash values are equal
        g1 = BGGenome("name1")
        g2 = BGGenome("name2")
        self.assertNotEqual(g1, g2)
        g2.name = "name1"
        self.assertEqual(g1, g2)
        self.assertNotEqual(g1, 5)
        self.assertNotEqual(g1, "name1")
        self.assertNotEqual(g1, [g1])

    def test_json_serialization_no_subclassing(self):
        # genome can be serialized into json format keeping all important information
        g = BGGenome("name1")
        ref_result = {
            "name": "name1",
            "g_id": g.json_id
        }
        self.assertDictEqual(g.to_json(schema_info=False), ref_result)
        ref_result[BGGenome_JSON_SCHEMA_JSON_KEY] = g.json_schema_name
        self.assertDictEqual(g.to_json(), ref_result)

    def test_json_deserialization_no_subclassing(self):
        # simple case
        json_object = {
            "name": "name1",
            "g_id": 1
        }
        result = BGGenome.from_json(data=json_object)
        self.assertEqual(result.name, "name1")
        # g_id is not mandatory for genome deserialization itself, but is required by the supervising class
        self.assertEqual(BGGenome.from_json(data={"name": "name1"}).name, "name1")
        # BGGenome scheme info shall be ignored at this level, as it is supplied by the supervising class
        self.assertEqual(BGGenome.from_json(data={"name": "name1",
                                                  BGGenome_JSON_SCHEMA_JSON_KEY: "lalal"}).name, "name1")
        # error case when "name" is not present
        with self.assertRaises(ValueError):
            BGGenome.from_json(data={})

    def test_json_deserialization_subclassing(self):
        # being provided an explicit JSONSchema, it shall be utilized for json deserialization
        class BGGenomeJSONSchemaNameOptional(BGGenome.BGGenomeJSONSchema):
            def make_object(self, data):
                if "name" not in data:
                    data["name"] = "default_name"
                return super().make_object(data=data)
        self.assertEqual(BGGenome.from_json(data={}, json_schema_class=BGGenomeJSONSchemaNameOptional).name, "default_name")

    def test__lt__(self):
        # genome is less than any non BGGenome instance
        # with other BGGenome instance it is compared by respective "name" attributes
        g1 = BGGenome("genome1")
        g2 = BGGenome("genome2")
        self.assertLess(g1, g2)
        self.assertGreater(g2, g1)
        g1 = BGGenome("genome1")
        g2 = BGGenome("genome")
        self.assertGreater(g1, g2)
        self.assertLess(g2, g1)
        # BGGenome is always smaller than non-BGGenome objects
        objects_to_compare_to = [1, (1,), [1], "a"]
        for object_to_compare_to in objects_to_compare_to:
            self.assertLess(g1, object_to_compare_to)
            self.assertLess(g2, object_to_compare_to)

    def test__le__(self):
        # Genome is considered less or equal to any other BGGenome is it is either less ("<" implementation
        # or equal (__eq__ implementation), than supplied argument
        g1 = BGGenome("genome1")
        g2 = BGGenome("genome1")
        self.assertLessEqual(g1, g2)
        self.assertLessEqual(g2, g1)
        self.assertTrue(g1 <= g2 <= g1)
        g3 = BGGenome("genome")
        self.assertLessEqual(g3, g1)
        self.assertLessEqual(g3, g1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from bisect import bisect_left
from marshmallow import Schema, fields

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

INFINITY_VERTEX_IDENTIFIER = "__infinity"
BGVertex_JSON_SCHEMA_JSON_KEY = "_py__bg_vertex_json_schema"


class BGVertex(object):
    """ An base class that represents a vertex (node) with all associated information in a breakpoint graph data structure

    While class represents a base inheritance point for specific vertex implementations, it does implement most of
    business logic operations, that vertex shall support.

    While different type of vertices are to be represented with different python classes, they all have a string representation,
    which mainly relies one the `name` attribute.
    """

    # this class based variable is utilized to the purpose if separating vertex root name from any additions that would specify some special properties of the vertex
    # (like vertex classes, special properties, etc)
    NAME_SEPARATOR = "__"

    # each vertex is json serializable/deserializable and this is implemented with a help of marshmallow library and respective manually defined schema
    class BGVertexJSONSchema(Schema):
        name = fields.String(required=True, attribute="name")
        v_id = fields.Int(required=True, attribute="json_id")
        _py__bg_vertex_json_schema = fields.String(attribute="json_schema_name")

        def make_object(self, data):
            try:
                return BGVertex(name=data["name"])
            except KeyError:
                raise ValueError("No `name` key in supplied json data for vertex deserialization")

    # a schema class based variable with json deserialization schema
    # must be updated in all heirs, as schema specific method `make_object` specifies the type of deserialized object
    json_schema = BGVertexJSONSchema()

    def __init__(self, name):
        self._name = None
        self.name = name

    @property
    def name(self):
        return str(self._name)

    @name.setter
    def name(self, value):
        self._name = value

    def __hash__(self):
        # all vertex are hashable objects and are uniquely defined by their name, thus a has value of vertex is just a hash value of its name
        return hash(self.name)

    def __eq__(self, other):
        # vertices are equal only if their class is equal as well as their names
        # in 99% of a time name is class specific and one can distinguish between vetices classes by only their names
        if not isinstance(other, self.__class__):
            return False
        return hash(self) == hash(other)

    def __getattr__(self, item):
        # this class serves as a stopper for all calls in form is_*_vertex, which is designed to test some combinatorial propoerties of vertex
        # base class does not belong to any specific group and thus answers `False` for all such calls
        # the rest of lookups is forwarded to the "next in mro chain"
        if item.startswith("is_") and item.endswith("_vertex"):
            return False
        return super().__getattribute__(item)

    @property
    def json_id(self):
        # a spicific vertex unique identifier that is utilized for aliasing vertices in json files (vertex might be references more than once), that is why
        # a reference to object by its json_id is used, rather than the full vertex object
        return hash(self)

    @property
    def json_schema_name(self):
        # each vertex is serialized with a help of some json schema (marshmallow powered), and information about such schema can be provided in
        # serialized vertex json object and further used to deserialized a vertex object into specific vertex class
        return self.json_schema.__class__.__name__

    def to_json(self, schema_info=True):
        # json serialization method that accounts for a possibility of excluding some schema specified fields from the result
        # also, since there are usually thousands of vertices serialized at the same time, no new schema object is created every time,
        # but rather some small monkey patching is performed with `exclude` field of json schema
        old_exclude_fields = self.json_schema.exclude
        new_exclude_fields = list(old_exclude_fields)
        if not schema_info:
            new_exclude_fields.append(BGVertex_JSON_SCHEMA_JSON_KEY)
        # monkey patch schema `exclude` attribute to ignore some fields in result json object
        self.json_schema.exclude = new_exclude_fields
        result = self.json_schema.dump(self).data
        # reverse the result of monkey patching
        self.json_schema.exclude = old_exclude_fields
        return result

    @classmethod
    def from_json(cls, data, json_schema_class=None):
        # deserialization from json is performed by internal machinery of `make_object` method, that is invoked transparently
        # specific json schema class can be specified and used for deserialization
        schema = cls.json_schema if json_schema_class is None else json_schema_class()
        return schema.load(data).data

    @staticmethod
    def get_vertex_class_from_vertex_name(string):
        # since every vertex even of different classes shall have a class specific `name` attribute, is is possible to distinguish between vertices classes
        # default value is BlockVertex, the most utilized vertex in the standard breakpoint graph
        if InfinityVertex.NAME_SUFFIX in string.split(BGVertex.NAME_SEPARATOR)[1:]:
            return TaggedInfinityVertex
        else:
            return TaggedBlockVertex

    @staticmethod
    def get_vertex_name_root(string):
        # as every vertex is represented by its name, some additional information about vertex class, type, etc. can be encoded into its name
        # such encoding is usually performed by appending some special suffixes to vertex name and utilizing `NAME_SEPARATOR` attribute
        return string.split(BGVertex.NAME_SEPARATOR)[0]


class BlockVertex(BGVertex):
    """ This class represents a special type of breakpoint graph vertex that correspond to a generic block extremity (gene/ synteny block/ etc.) """

    class BlockVertexJSONSchema(BGVertex.BGVertexJSONSchema):
        """ JSON schema for this class is redefined to tune the `make_object` method, that shall return `BlockVertex` instance, rather than `BGVertex` one """
        def make_object(self, data):
            try:
                return BlockVertex(name=data["name"])
            except KeyError:
                raise ValueError("No `name` key in supplied json data for vertex deserialization")

    # a new JSON schema is initialized and set of be used for all instance of `VertexClass`
    json_schema = BlockVertexJSONSchema()

    def __init__(self, name, mate_vertex=None):
        super().__init__(name=name)
        self.mate_vertex = mate_vertex

    @property
    def is_regular_vertex(self):
        """ This class implements a property check for vertex to belong to class of regular vertices """
        return True

    @property
    def block_name(self):
        if self.is_block_vertex:
            if self.is_tail_vertex or self.is_head_vertex:
                return self._name[:-1]
            else:
                return self._name

    @property
    def is_block_vertex(self):
        """ This class implements a property check for vertex to belong to a class of vertices, that correspond to extremities of genomic blocks"""
        return True

    @property
    def is_head_vertex(self):
        return self.is_block_vertex and self._name.endswith("h")

    @property
    def is_tail_vertex(self):
        return self.is_block_vertex and self._name.endswith("t")

    @classmethod
    def from_json(cls, data, json_schema_class=None):
        """ This class overwrites the from_json method thus, making sure, that if `from_json` is called from this class instance, it will provide its JSON schema as a default one """
        json_schema = cls.json_schema if json_schema_class is None else json_schema_class()
        return super().from_json(data=data, json_schema_class=json_schema.__class__)


class InfinityVertex(BGVertex):
    """ This class represents a special type of breakpoint graph vertex that correspond to a generic extremity of genomic fragment (chromosome, scaffold, contig, etc.)"""

    class InfinityVertexJSONSchema(BGVertex.BGVertexJSONSchema):
        """ JSON Schema for this class is redefined to tune the `make_object` method, that shall return `InfinityVertex` instance, rather than a `BGVertex` one """
        def make_object(self, data):
            try:
                json_name = data["name"]
                name = json_name.split(InfinityVertex.NAME_SUFFIX)[0]
                return InfinityVertex(name=name)
            except KeyError:
                raise ValueError("No `name` key in supplied json data for vertex deserialization")

    # InfinityVertex instances have a special suffix in their name that is determined by a class variable `NAME_SUFFIX`
    NAME_SUFFIX = "infinity"

    # a setup for a new JSON schema is performed class-wise to be utilized by all instance of InfinityVertex
    json_schema = InfinityVertexJSONSchema()

    def __init__(self, name):
        # current class allows for a standard access to the `name` attribute, but performs transparent computation behind the scenes
        # so the name is stored in a private variable __name, and property `name` is implemented
        super().__init__(name=name)

    @property
    def name(self):
        """ access to classic name attribute is hidden by this property """
        return self.NAME_SEPARATOR.join([super().name, self.NAME_SUFFIX])

    @name.setter
    def name(self, value):
        """When someone wants to set a new name for the `InfinityVertex` instance, it is transparently store into the `__name` attribute """
        self._name = value

    @property
    def is_irregular_vertex(self):
        """ This class implements a property check for vertex to belong to a class of vertices, that correspond to extremities of genomic fragments """
        return True

    @property
    def is_infinity_vertex(self):
        """ This class implements a property check for vertex to belong to a class of vertices, that correspond to standard extremities of genomic fragments """
        return True

    @classmethod
    def from_json(cls, data, json_schema_class=None):
        """ This class overwrites the from_json method, thus making sure that if `from_json` is called from this class instance, it will provide its JSON schema as a default one"""
        schema = cls.json_schema if json_schema_class is None else json_schema_class()
        return super().from_json(data=data, json_schema_class=schema.__class__)


class TaggedVertex(BGVertex):

    class TaggedVertexJSONSchema(BGVertex.BGVertexJSONSchema):

        def make_object(self, data):
            predefined_object_class = getattr(self, "object_class", None)
            object_class = TaggedVertex if predefined_object_class is None else predefined_object_class
            try:
                json_name = data["name"]
                split_name = json_name.split(TaggedVertex.NAME_SEPARATOR)
                root = split_name[0]
                tags = list(filter(lambda name_part: TaggedVertex.TAG_SEPARATOR in name_part, split_name[1:]))
                tags = [entry.split(TaggedVertex.TAG_SEPARATOR) for entry in tags]
                result = object_class(name=root)
                result.tags = sorted(tags)
                return result
            except KeyError:
                raise ValueError("No `name` key in supplied json data for vertex deserialization")

    TAG_SEPARATOR = ":"

    json_schema = TaggedVertexJSONSchema()

    def __init__(self, name):
        self.tags = []
        super().__init__(name=name)

    @property
    def is_tagged_vertex(self):
        return True

    @property
    def name(self):
        """ access to classic name attribute is hidden by this property """
        return self.NAME_SEPARATOR.join([super().name] +
                                        [self.TAG_SEPARATOR.join([str(tag), str(value)]) for tag, value in self.tags])

    @name.setter
    def name(self, value):
        """ shared "protected" variable for the name storing attribute """
        self._name = value

    def add_tag(self, tag, value):
        """ as tags are kept in a sorted order, a bisection is a fastest way to identify a correct position
        of or a new tag to be added. An additional check is required to make sure w don't add duplicates
        """
        index = bisect_left(self.tags, (tag, value))
        contains = False
        if index < len(self.tags):
            contains = self.tags[index] == (tag, value)
        if not contains:
            self.tags.insert(index, (tag, value))

    def __getattr__(self, item):
        """  """
        if item.startswith("is_") and item.endswith("_vertex"):
            tag = item[3:-7]
            index = bisect_left([tag_name for tag_name, _ in self.tags], tag)
            if index < len(self.tags):
                return self.tags[index][0] == tag
        return super().__getattr__(item)

    def remove_tag(self, tag, value, silent_fail=False):
        """ we try to remove supplied pair tag -- value, and if does not exist outcome depends on the silent_fail flag """
        try:
            self.tags.remove((tag, value))
        except ValueError as err:
            if not silent_fail:
                raise err

    @classmethod
    def from_json(cls, data, json_schema_class=None):
        """ This class overwrites the from_json method, thus making sure that if `from_json` is called from this class instance, it will provide its JSON schema as a default one"""
        schema = cls.json_schema if json_schema_class is None else json_schema_class()
        return super().from_json(data=data, json_schema_class=schema.__class__)

class TaggedBlockVertex(BlockVertex, TaggedVertex):
    class TaggedBlockVertexJSONSchema(TaggedVertex.TaggedVertexJSONSchema, BlockVertex.BlockVertexJSONSchema):
        def make_object(self, data):
            setattr(self, "object_class", TaggedBlockVertex)
            return super().make_object(data)

    json_schema = TaggedBlockVertexJSONSchema()

class TaggedInfinityVertex(InfinityVertex, TaggedVertex):
    class TaggedInfinityVertexJSONSchema(TaggedVertex.TaggedVertexJSONSchema, InfinityVertex.InfinityVertexJSONSchema):
        def make_object(self, data):
            setattr(self, "object_class", TaggedInfinityVertex)
            return super().make_object(data)

    json_schema = TaggedInfinityVertexJSONSchema()




# -*- coding: utf-8 -*-
from collections import Counter


__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"


class KBreak(object):
    """ A generic object that can represent any k-break ( k>= 2)

    A notion of k-break arises from the bioinformatics combinatorial object BreakpointGraph and is first mentioned in http://home.gwu.edu/~maxal/ap_tcs08.pdf
    A generic k-break operates on k specified edges of spisific multicolor and replaces them with another set of k edges with the same multicolor on the same set of vertices in way, that the degree of vertices is kept intact.

    Initialization of the instance of :class:`KBreak` is performed with a validity check of supplied data, which must comply with the definition of k-break.

    Class carries following attributes carrying information about k-break structure:

    * :attr:`KBreak.start_edges`: a list of edges (in terms of paired vertices) that are to be removed by current :class:`KBreak`
    * :attr:`KBreak.result_edges`: a list of edges (in terms of paired vertices) that are to be created by current :class:`KBreak`
    * :attr:`KBreak.multicolor`: a :class:`bg.multicolor.Multicolor` instance, that specifies the multicolor of edges that are to be removed / created by current :class:`KBreak`

    Main operations:

    * :meth:`KBreak.valid_kbreak_matchings`: a method that checks if provided sets of started / resulted edges comply with the notions ob k-break definition
    """
    def __init__(self, start_edges, result_edges, multicolor):
        """ Initialization of :class:`KBreak` object.

        The initialization process consists of multiple checks, before any assignment and initialization itself is performed.

        First checks the fact, that information about start / result edges is supplied in form of paired vertices.
        Then check is performed to make sure, that degrees of vertices, that current :class:`KBreak` operates on, is preserved.

        :param start_edges: a list of pairs of vertices, that specifies where edges shall be removed by current :class:`KBreak`
        :type start_edges: ``list(tuple(vertex, vertex), ...)``
        :param result_edges: a list of pairs of vertices, that specifies where edges shall be created by current :class:`KBreak`
        :type result_edges: ``list(tuple(vertex, vertex), ...)``
        :param multicolor: a multicolor, that specifies which edges between specified pairs of vertices are to be removed / created
        :type multicolor: :class:`bg.multicolor.Multicolor`
        :return: a new instance of :class:`Kbreak`
        :rtype: :class:`KBreak`
        :raises: ``ValueError``
        """
        self.start_edges = start_edges
        self.result_edges = result_edges
        self.multicolor = multicolor
        for vertex_pair in self.start_edges:
            if len(vertex_pair) != 2:
                raise ValueError("Expected edges in a form of pairs of vertices.\n "
                                 "Not a pair of vertices ({issue}) in start edges."
                                 "".format(issue=str(vertex_pair)))
        for vertex_pair in self.result_edges:
            if len(vertex_pair) != 2:
                raise ValueError("Expected edges in a form of pairs of vertices.\n "
                                 "Not a pair of vertices ({issue}) in result edges."
                                 "".format(issue=str(vertex_pair)))
        if not KBreak.valid_kbreak_matchings(start_edges=self.start_edges,
                                             result_edges=self.result_edges):
            raise ValueError("Supplied sets of start and result edges do not correspond to "
                             "correct k-break operation (either the set of vertices is not consistent, or "
                             "the degrees of vertices change)")

    @staticmethod
    def valid_kbreak_matchings(start_edges, result_edges):
        """ A staticmethod check implementation that makes sure that degrees of vertices, that are affected by current :class:`KBreak`

        By the notion of k-break, it shall keep the degree of vertices in :class:`bg.breakpoint_graph.BreakpointGraph` the same, after its application.
        By utilizing the Counter class, such check is performed, as the number the vertex is mentioned corresponds to its degree.

        :param start_edges: a list of pairs of vertices, that specifies where edges shall be removed by :class:`KBreak`
        :type start_edges: ``list(tuple(vertex, vertex), ...)``
        :param result_edges: a list of pairs of vertices, that specifies where edges shall be created by :class:`KBreak`
        :type result_edges: ``list(tuple(vertex, vertex), ...)``
        :return: a flag indicating if the degree of vertices are equal in start / result edges, targeted by :class:`KBreak`
        :rtype: ``Boolean``
        """
        start_stats = Counter(vertex for vertex_pair in start_edges for vertex in vertex_pair)
        result_stats = Counter(vertex for vertex_pair in result_edges for vertex in vertex_pair)
        return start_stats == result_stats


# -*- coding: utf-8 -*-
from unittest.mock import Mock

__author__ = "Sergey Aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "production"

import unittest
from bg.multicolor import Multicolor
from bg.genome import BGGenome


class MulticolorTestCase(unittest.TestCase):
    # in comments bellow "genome" and "color" represent the same concept
    # if not specified otherwise explicitly

    def setUp(self):
        # some heavily used variables
        self.genome1 = BGGenome("red")
        self.genome2 = BGGenome("green")
        self.genome3 = BGGenome("blue")
        self.genome4 = BGGenome("black")
        self.genome5 = BGGenome("yellow")

    def test_empty_initialization(self):
        # empty multicolor shall be initialisable but contain no information about the colors or their multiplicity
        mc = Multicolor()
        self.assertEqual(len(mc.colors), 0)
        self.assertEqual(len(mc.multicolors), 0)

    def test_single_initialization(self):
        # simple case initialization where only one genome with multiplicity one is supplied
        mc = Multicolor(self.genome1)
        self.assertEqual(len(mc.colors), 1)
        self.assertEqual(len(mc.multicolors), 1)
        self.assertSetEqual({self.genome1}, mc.colors)
        self.assertEqual(mc.multicolors[self.genome1], 1)
        mc = Multicolor(self.genome1)
        self.assertEqual(len(mc.colors), 1)
        self.assertEqual(len(mc.multicolors), 1)
        self.assertSetEqual({self.genome1}, mc.colors)
        self.assertEqual(mc.multicolors[self.genome1], 1)

    def test_multiple_initialization(self):
        # cases when multiple genomes with different multiplicities (from 1 to >1 are specified)
        mc = Multicolor(self.genome1, self.genome2, self.genome3)
        self.assertEqual(len(mc.colors), 3)
        self.assertEqual(len(mc.multicolors), 3)
        self.assertSetEqual({self.genome1, self.genome2, self.genome3}, mc.colors)
        for color in mc.multicolors:
            self.assertEqual(mc.multicolors[color], 1)
        mc = Multicolor(*[self.genome1, self.genome2, self.genome3])
        self.assertEqual(len(mc.colors), 3)
        self.assertEqual(len(mc.multicolors), 3)
        self.assertSetEqual({self.genome1, self.genome2, self.genome3}, mc.colors)
        for color in mc.multicolors:
            self.assertEqual(mc.multicolors[color], 1)
        mc1 = Multicolor(self.genome1, self.genome2, self.genome1)
        self.assertEqual(len(mc1.colors), 2)
        self.assertEqual(mc1.multicolors[self.genome2], 1)
        self.assertEqual(mc1.multicolors[self.genome1], 2)

    def test_equality(self):
        # multicolors are called equal
        # if they contain information about hte same colors with same multiplicity for each color
        mc1 = Multicolor(self.genome1)
        mc2 = Multicolor(self.genome1)
        self.assertEqual(mc1, mc2)
        mc1 = Multicolor(self.genome1, self.genome2)
        mc2 = Multicolor(self.genome2, self.genome1)
        self.assertEqual(mc1, mc2)
        mc1 = Multicolor(self.genome1)
        mc2 = Multicolor(BGGenome("ret"))
        self.assertNotEqual(mc1, mc2)
        mc1 = Multicolor(self.genome1, self.genome1, self.genome3)
        mc2 = Multicolor(self.genome1, self.genome3)
        self.assertNotEqual(mc1, mc2)
        ###############################################################################################
        #
        # cases when Multicolor object is compared with a non-Multicolor object
        # this equality comparison is always False
        #
        ###############################################################################################
        for non_multicolor_type_object in [1, (1,), [1], "1", Mock()]:
            self.assertNotEqual(Multicolor(), non_multicolor_type_object)

    def test_update(self):
        # multicolor can be updated by multiple arguments
        # they shall add information about colors (if color was not present before) and/or their multiplicity
        # change is inplace
        mc = Multicolor()
        self.assertSetEqual(set(), mc.colors)
        mc.update(self.genome1)
        self.assertEqual(len(mc.colors), 1)
        self.assertEqual(len(mc.multicolors), 1)
        self.assertSetEqual({self.genome1}, mc.colors)
        mc.update(self.genome2, self.genome3)
        self.assertEqual(len(mc.colors), 3)
        self.assertEqual(len(mc.multicolors), 3)
        self.assertSetEqual({self.genome1, self.genome2, self.genome3}, mc.colors)
        for color in mc.multicolors:
            self.assertEqual(mc.multicolors[color], 1)
        mc.update(self.genome1)
        self.assertEqual(len(mc.colors), 3)
        self.assertEqual(len(mc.multicolors), 3)
        for color in mc.multicolors:
            if color == self.genome1:
                self.assertEqual(mc.multicolors[color], 2)
            else:
                self.assertEqual(mc.multicolors[color], 1)
        self.assertSetEqual({self.genome1, self.genome2, self.genome3}, mc.colors)

    def test_left_merge(self):
        # multicolors can be merged
        # left_merge adds information about colors and their multiplicity from right specified genome, to the left one
        mc1 = Multicolor(self.genome1)
        mc2 = Multicolor(self.genome3)
        Multicolor.left_merge(mc1, mc2)
        self.assertEqual(len(mc1.colors), 2)
        self.assertEqual(len(mc1.multicolors), 2)
        for color in mc1.multicolors:
            self.assertEqual(mc1.multicolors[color], 1)
        self.assertSetEqual({self.genome1, self.genome3}, mc1.colors)
        self.assertEqual(len(mc2.colors), 1)
        self.assertEqual(len(mc2.multicolors), 1)
        self.assertSetEqual({self.genome3}, mc2.colors)
        for color in mc2.multicolors:
            self.assertEqual(mc2.multicolors[color], 1)
        mc3 = Multicolor.left_merge(Multicolor(), mc2)
        self.assertEqual(mc3, mc2)
        mc4 = Multicolor.left_merge(mc2, Multicolor())
        self.assertEqual(mc4, mc2)

    def test_merge(self):
        # multicolors can be merged into a new multicolor
        # "merge" method creates a new multicolor with information from a variable number of provided multicolors
        mc1 = Multicolor(self.genome1)
        mc2 = Multicolor(self.genome3)
        mc3 = Multicolor.merge(mc1, mc2)
        self.assertNotEqual(mc1, mc3)
        self.assertNotEqual(mc2, mc3)
        self.assertEqual(len(mc3.colors), 2)
        self.assertEqual(len(mc3.multicolors), 2)
        for color in mc3.multicolors:
            self.assertEqual(mc3.multicolors[color], 1)
        self.assertSetEqual({self.genome1, self.genome3}, mc3.colors)
        mc4 = Multicolor.merge(Multicolor(), mc2)
        self.assertEqual(mc2, mc4)
        mc5 = Multicolor.merge(mc1, mc2, mc3)
        self.assertNotEqual(mc5, mc3)
        self.assertEqual(len(mc5.colors), 2)
        for color in mc5.multicolors:
            self.assertEqual(mc5.multicolors[color], 2)
        mc6 = Multicolor.merge(mc1)
        self.assertEqual(mc1, mc6)

    def test_inplace_delete(self):
        # multicolor supports removing information about colors, taking into account to their multiplicity
        # multicolor support deletion using information from Multicolor instance or any iterable
        # negative counts for the color multiplicity are not supported:
        #   so if color had multiplicity 1 and this color was supposed to be deleted twice, it will be just deleted
        mc1 = Multicolor(self.genome1, self.genome3, self.genome2)
        mc1.delete((self.genome1,))
        self.assertTrue(self.genome1 not in mc1.colors)
        self.assertEqual(len(mc1.colors), 2)
        self.assertEqual(len(mc1.multicolors), 2)
        self.assertSetEqual({self.genome3, self.genome2}, mc1.colors)
        for color in mc1.multicolors:
            self.assertEqual(mc1.multicolors[color], 1)
        mc1.delete((self.genome5,))
        self.assertTrue(self.genome1 not in mc1.colors)
        self.assertEqual(len(mc1.colors), 2)
        self.assertSetEqual({self.genome3, self.genome2}, mc1.colors)
        self.assertEqual(len(mc1.multicolors), 2)
        for color in mc1.multicolors:
            self.assertEqual(mc1.multicolors[color], 1)
        mc2 = Multicolor(self.genome1, self.genome2, self.genome3)
        mc3 = Multicolor(self.genome1, self.genome2, self.genome5)
        mc2.delete(mc3)
        self.assertEqual(len(mc2.colors), 1)
        self.assertEqual(len(mc2.multicolors), 1)
        self.assertSetEqual({self.genome3}, mc2.colors)
        for color in mc2.multicolors:
            self.assertEqual(mc2.multicolors[color], 1)
        mc4 = Multicolor(self.genome1, self.genome1, self.genome2, self.genome3)
        mc5 = Multicolor(self.genome1, self.genome2, self.genome5)
        mc4.delete(mc5)
        self.assertEqual(len(mc4.colors), 2)
        self.assertEqual(len(mc4.multicolors), 2)
        self.assertSetEqual({self.genome3, self.genome1}, mc4.colors)
        for color in mc4.multicolors:
            self.assertEqual(mc4.multicolors[color], 1)

    def test__sub__(self):
        # - operator os overloaded to provide "delete" alike behaviour, but with a creation of a new Multicolor instance
        # only Multicolor instance is supported as an argument
        # for any other argument type a TypeError is raised
        mc1 = Multicolor(self.genome1, self.genome3, self.genome1, self.genome2)
        mc2 = Multicolor(self.genome3, self.genome2, self.genome5)
        mc3 = mc1 - mc2
        self.assertEqual(len(mc3.colors), 1)
        self.assertEqual(len(mc3.multicolors), 1)
        self.assertEqual(mc3.multicolors[self.genome1], 2)
        self.assertSetEqual({self.genome1}, mc3.colors)
        mc4 = Multicolor() - mc2
        self.assertEqual(mc4, Multicolor())
        with self.assertRaises(TypeError):
            mc1 - 5

    def test__isub__(self):
        # -= operator is overloaded and support only Multicolor instance as an argument
        # for any other argument a TypeError is raised
        # behalves just like the "delete" method
        mc1 = Multicolor(self.genome1, self.genome3, self.genome1, self.genome2)
        mc2 = Multicolor(self.genome3, self.genome2, self.genome5)
        mc1_id = id(mc1)
        mc1 -= mc2
        self.assertEqual(len(mc1.colors), 1)
        self.assertEqual(len(mc1.multicolors), 1)
        self.assertEqual(mc1.multicolors[self.genome1], 2)
        self.assertSetEqual({self.genome1}, mc1.colors)
        self.assertEqual(id(mc1), mc1_id)
        with self.assertRaises(TypeError):
            mc1 -= 5

    def test__add__(self):
        # + operator is overloaded and works just like a "merge" method, but support only Multicolor instance as an argument
        # for any other type of argument a TypeError is raised
        mc1 = Multicolor(self.genome1, self.genome2)
        mc2 = Multicolor(self.genome3, self.genome5, self.genome1)
        mc3 = mc1 + mc2
        self.assertEqual(len(mc3.colors), 4)
        self.assertEqual(len(mc3.multicolors), 4)
        self.assertSetEqual({self.genome1, self.genome2, self.genome3, self.genome5}, mc3.colors)
        for color in mc3.multicolors:
            if color == self.genome1:
                self.assertEqual(mc3.multicolors[color], 2)
            else:
                self.assertEqual(mc3.multicolors[color], 1)
        with self.assertRaises(TypeError):
            mc1 + 5

    def test__iadd__(self):
        # += operator is overloaded and works just like a "left_merge" method
        # but supports only a Multicolor instance as an argument
        # for any other type of argument a TypeError is raised
        mc1 = Multicolor(self.genome1, self.genome2)
        mc2 = Multicolor(self.genome3, self.genome5, self.genome1)
        mc1_id = id(mc1)
        mc1 += mc2
        self.assertEqual(len(mc1.colors), 4)
        self.assertEqual(len(mc1.multicolors), 4)
        self.assertSetEqual({self.genome1, self.genome2, self.genome3, self.genome5}, mc1.colors)
        for color in mc1.multicolors:
            if color == self.genome1:
                self.assertEqual(mc1.multicolors[color], 2)
            else:
                self.assertEqual(mc1.multicolors[color], 1)
        self.assertEqual(mc1_id, id(mc1))
        with self.assertRaises(TypeError):
            mc1 += 5

    def test__lt__and__le__(self):
        # multicolor are compared as follows:
        # for all the colors in the left argument of comparison, checks that multiplicity of that color in right argument is
        # less (less-equal)
        mc1 = Multicolor(self.genome1, self.genome2, self.genome1)
        mc2 = Multicolor(self.genome1, self.genome2)
        self.assertTrue(mc2 < mc1)
        self.assertTrue(mc2 <= mc1)
        self.assertFalse(mc2 <= 5)
        self.assertFalse(mc2 < 5)
        mc2 = Multicolor(self.genome1, self.genome1, self.genome2)
        self.assertFalse(mc2 < mc1)
        self.assertTrue(mc2 <= mc1)
        mc2 = Multicolor(self.genome1, self.genome1, self.genome2, self.genome2)
        self.assertFalse(mc2 < mc1)
        self.assertFalse(mc2 <= mc1)

    def test__gt__and__ge__(self):
        # multicolor are compared as follows:
        # for all the colors in the left argument of comparison, checks that multiplicity of that color in right argument is
        # greater (greater-equal)
        mc1 = Multicolor(self.genome1, self.genome2, self.genome1)
        mc2 = Multicolor(self.genome1, self.genome2)
        self.assertTrue(mc1 > mc2)
        self.assertTrue(mc1 >= mc2)
        mc2 = Multicolor(self.genome1, self.genome1, self.genome2)
        self.assertFalse(mc1 > mc2)
        self.assertTrue(mc1 >= mc2)
        mc2 = Multicolor(self.genome1, self.genome1, self.genome2, self.genome2)
        self.assertFalse(mc1 > mc2)
        self.assertFalse(mc1 >= mc2)
        ###############################################################################################
        #
        # Multicolor object is never greater or equal to the non-Multicolor object
        #
        ###############################################################################################
        for non_multicolor_object in [1, (1,), [1,], "1", Mock()]:
            self.assertFalse(mc1 >= non_multicolor_object)
            self.assertFalse(mc1 > non_multicolor_object)

    def test_similarity_score(self):
        # similarity score for multicolors is computed as follows:
        # for every mutual color, the smallest multiplicity of such color in two Multicolors is considered
        # similarity score is computed as the sum over such min multiplicities of all shared colors
        mc1 = Multicolor(self.genome1, self.genome2)
        mc2 = Multicolor(self.genome1, self.genome2)
        self.assertEqual(Multicolor.similarity_score(mc1, mc2), 2)
        self.assertEqual(Multicolor.similarity_score(mc2, mc1), 2)
        mc2 = Multicolor(self.genome1, self.genome1)
        self.assertEqual(Multicolor.similarity_score(mc1, mc2), 1)
        self.assertEqual(Multicolor.similarity_score(mc2, mc1), 1)
        mc2 = Multicolor(self.genome1)
        self.assertEqual(Multicolor.similarity_score(mc1, mc2), 1)
        self.assertEqual(Multicolor.similarity_score(mc2, mc1), 1)
        mc2 = Multicolor(self.genome4)
        self.assertEqual(Multicolor.similarity_score(mc1, mc2), 0)
        self.assertEqual(Multicolor.similarity_score(mc2, mc1), 0)

    ###############################################################################################
    #
    # following tests are aimed at generic splitting based on provided set of good colors
    # it is compatible with tree consistent multicolor splitting, but is not limited to it
    #
    ###############################################################################################

    def test_split_colors_account_for_multiplicity_in_guidance(self):
        ###############################################################################################
        #
        # when no guidance is specified, a multicolor shall be split according to its own colors
        # when `account_for_multiplicity_in_guidance` is specified
        #   each color in the splitted result shall have multiplicity as it had in the targeted multicolor
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome3, self.genome3, self.genome3)
        ref = [Multicolor(self.genome1, self.genome1),
               Multicolor(self.genome2),
               Multicolor(self.genome3, self.genome3, self.genome3)]
        result = Multicolor.split_colors(mc, account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 3)
        for result_mc in result:
            self.assertIn(result_mc, ref)
        ###############################################################################################
        ###############################################################################################
        #
        # a simple guidance with a single multicolor, that has only a single color with multiplicity one
        #
        ###############################################################################################
        mc = Multicolor(self.genome1)
        guidance = [Multicolor(self.genome1)]
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 1)
        mc = result[0]
        self.assertEqual(len(mc.colors), 1)
        self.assertEqual(len(mc.multicolors), 1)
        self.assertEqual(mc.multicolors[self.genome1], 1)
        ###############################################################################################
        ###############################################################################################
        #
        # color exists in guidance only as subset, then it shall be retrieved fully on its own
        #
        ###############################################################################################
        mc = Multicolor(self.genome1)
        guidance = [Multicolor(self.genome1, self.genome4)]
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 1)
        mc = result[0]
        self.assertEqual(len(mc.colors), 1)
        self.assertEqual(len(mc.multicolors), 1)
        self.assertEqual(mc.multicolors[self.genome1], 1)
        ###############################################################################################
        ###############################################################################################
        #
        # color exists in guidance both as subset and a set itself, and thus shall be retrieved fully
        #
        ###############################################################################################
        mc = Multicolor(self.genome1)
        guidance = [Multicolor(self.genome1), Multicolor(self.genome1, self.genome4), Multicolor(self.genome4)]
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 1)
        mc = result[0]
        self.assertEqual(len(mc.colors), 1)
        self.assertEqual(len(mc.multicolors), 1)
        self.assertEqual(mc.multicolors[self.genome1], 1)
        ###############################################################################################
        ###############################################################################################
        #
        # color does not exist in guidance, and shall be retrieved fully, as an appendix
        #
        ###############################################################################################
        mc = Multicolor(self.genome1)
        guidance = [Multicolor(self.genome2, self.genome4)]
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 1)
        mc = result[0]
        self.assertEqual(len(mc.colors), 1)
        self.assertEqual(len(mc.multicolors), 1)
        self.assertEqual(mc.multicolors[self.genome1], 1)
        ###############################################################################################
        ###############################################################################################
        #
        # some color in guidance present twice in the splitting multicolor
        # and thus shall be retrieved fully twice
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome1, self.genome2, self.genome2)
        guidance = [Multicolor(self.genome1, self.genome2)]
        ref1 = guidance[0]
        ref2 = Multicolor(self.genome1)
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 3)
        for result_mc in result:
            self.assertIn(result_mc, [ref1, ref2])
        ###############################################################################################
        ###############################################################################################
        #
        # some colors in guidance have non empty intersections (with multiplicity > 1) with s splitting color
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome2)
        guidance = [Multicolor(self.genome1, self.genome2, self.genome3)]
        ref = Multicolor(self.genome1, self.genome2)
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 2)
        for result_mc in result:
            self.assertEqual(result_mc, ref)
        ###############################################################################################
        ###############################################################################################
        #
        # some color in guidance is present twice in the splitting color
        # some color in guidance has a non empty intersection with splitting color
        #   that interferes with multicolor in guidance, that is fully present
        # full presence must overtake in this case
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome2)
        guidance = [Multicolor(self.genome1, self.genome2), Multicolor(self.genome1, self.genome2, self.genome3)]
        ref = Multicolor(self.genome1, self.genome2)
        result = Multicolor.split_colors(mc, guidance=guidance,
                                         account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 2)
        for result_mc in result:
            self.assertEqual(result_mc, ref)
        ###############################################################################################
        ###############################################################################################
        #
        # both fully present and non-empty intersection colors are present in guidance
        # the most complex test case
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome2, self.genome3)
        guidance = [Multicolor(self.genome1, self.genome2, self.genome3),
                    Multicolor(self.genome1, self.genome1, self.genome2, self.genome2)]
        ref1 = Multicolor(self.genome1, self.genome1, self.genome2, self.genome2)
        ref2 = Multicolor(self.genome3)
        result = Multicolor.split_colors(mc, guidance=guidance,
                                         account_for_color_multiplicity_in_guidance=True)
        self.assertEqual(len(result), 2)
        for result_mc in result:
            self.assertIn(result_mc, [ref1, ref2])

    def test_split_colors_do_not_account_for_multiplicity_in_guidance(self):
        ###############################################################################################
        ###############################################################################################
        #
        # no guidance, targeted multicolor shall be split on separate colors
        # keeping respective colors multiplicity intact in each splitted peace
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome3, self.genome3, self.genome3)
        result = Multicolor.split_colors(mc, account_for_color_multiplicity_in_guidance=False)
        ref = [Multicolor(self.genome1, self.genome1),
               Multicolor(self.genome2),
               Multicolor(self.genome3, self.genome3, self.genome3)]
        self.assertEqual(len(result), 3)
        for result_mc in result:
            self.assertIn(result_mc, ref)
        ###############################################################################################
        ###############################################################################################
        #
        # simple case, where guidance contains already multicolor with multiplicity 1
        # targeted multicolor shall be split based on those colors
        #     but multiplicity of respective colors in the result shall be kept as it was in the targetted multicolor
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome3, self.genome3, self.genome3)
        guidance = [Multicolor(self.genome1, self.genome2), Multicolor(self.genome3)]
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=False)
        ref = [Multicolor(self.genome1, self.genome1, self.genome2),
               Multicolor(self.genome3, self.genome3, self.genome3)]
        self.assertEqual(len(result), 2)
        for result_mc in result:
            self.assertIn(result_mc, ref)
        ###############################################################################################
        ###############################################################################################
        #
        # more complex case, when guidance contains multicolor with multiplicity of colors bigger than 1
        # in this case, those guidance multicolors will be simplified to same colors multicolors
        #   but multiplicity of respective colors will be changed to 1
        # resulted multicolor split shall contain multiplicity of respective colors, as in the original one
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome3, self.genome3, self.genome3)
        guidance = [Multicolor(self.genome1, self.genome2, self.genome1), Multicolor(self.genome3, self.genome3)]
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=False)
        ref = [Multicolor(self.genome1, self.genome1, self.genome2),
               Multicolor(self.genome3, self.genome3, self.genome3)]
        self.assertEqual(len(result), 2)
        for result_mc in result:
            self.assertIn(result_mc, ref)
        ###############################################################################################
        ###############################################################################################
        #
        # case when guidance has multiple multicolors, that after simplification would look the same
        #   (they differ only in the multiplicity of respective colors)
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome2, self.genome3, self.genome3, self.genome4, self.genome4)
        guidance = [Multicolor(self.genome1), Multicolor(self.genome1, self.genome1), Multicolor(self.genome1, self.genome1),
                    Multicolor(self.genome2, self.genome3),
                    Multicolor(self.genome4)]
        result = Multicolor.split_colors(mc, guidance=guidance, account_for_color_multiplicity_in_guidance=False)
        ref = [Multicolor(self.genome1),
               Multicolor(self.genome2, self.genome3, self.genome3),
               Multicolor(self.genome4, self.genome4)]
        self.assertEqual(len(result), 3)
        for result_mc in result:
            self.assertIn(result_mc, ref)

    def test_color_splits_no_guidance_sorting(self):
        ###############################################################################################
        #
        # order of multicolors in guidance affects the splitting
        # if two colors in guidance are both present in the splitting multicolor
        #   then the first multicolor in the guidance  will be retrieved, but the second might not be,
        #   as not enough information will be left in the splitting multicolor
        #
        ###############################################################################################
        ###############################################################################################
        #
        # simple case, when there are two multicolors in the guidance
        # we don't account for the
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome2, self.genome3)
        guidance = [Multicolor(self.genome1, self.genome2),
                    Multicolor(self.genome1, self.genome2, self.genome3)]
        result = Multicolor.split_colors(mc, guidance=guidance, sorted_guidance=True)
        self.assertEqual(len(result), 2)
        ref = [Multicolor(self.genome1, self.genome2), Multicolor(self.genome3)]
        for result_mc in result:
            self.assertIn(result_mc, ref)
        ###############################################################################################
        ###############################################################################################
        #
        # simple case, when there are two multicolors in the guidance
        # we don't account for the
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome1, self.genome2, self.genome3, self.genome3, self.genome3)
        guidance = [Multicolor(self.genome1),
                    Multicolor(self.genome1, self.genome2),
                    Multicolor(self.genome3, self.genome3, self.genome3, self.genome2)]
        result = Multicolor.split_colors(mc, guidance=guidance, sorted_guidance=True)
        ref = [Multicolor(self.genome1), Multicolor(self.genome2, self.genome3, self.genome3, self.genome3)]
        self.assertEqual(len(result), 3)
        for result_mc in result:
            self.assertIn(result_mc, ref)

    def test_split_colors_with_empty_multicolor_in_guidance(self):
        ###############################################################################################
        #
        # empty multicolor in splitting guidance shall not have any affect on the splitting procedure
        #
        ###############################################################################################
        mc = Multicolor(self.genome1, self.genome2)
        guidance = [Multicolor(self.genome1), Multicolor()]
        result = Multicolor.split_colors(mc, guidance=guidance)
        self.assertEqual(len(result), 2)
        ref = [Multicolor(self.genome1), Multicolor(self.genome2)]
        for result_mc in result:
            self.assertIn(result_mc, ref)

    ###############################################################################################
    #
    # end of splitting test cases
    #
    ###############################################################################################

    def test_hashable_representation(self):
        # every multicolor has to have a hashable representation, that can be utilized in a set/dict
        # for a fast check against multicolor instance
        ################################################
        # the idea is to use sorted Counter.elements() method and convert into sorted tuple on the fly
        genome_list = [self.genome1, self.genome2, self.genome3, self.genome4, self.genome1, self.genome2, self.genome1]
        mc = Multicolor(*genome_list)
        ref_tuple = tuple(sorted(genome_list))
        result = mc.hashable_representation
        self.assertTrue(isinstance(result, tuple))
        self.assertTupleEqual(result, ref_tuple)
        mc1 = Multicolor(*result)
        self.assertEqual(mc, mc1)
        # non-equal multicolors shall have different hashable representations
        mc1 = Multicolor(*genome_list[:-2])
        mc2 = Multicolor(*genome_list[:-1])
        self.assertNotEqual(mc1, mc2)
        self.assertNotEqual(mc1.hashable_representation, mc2.hashable_representation)
        # there shall be no errors or exceptions raised while taking hash of hashable_representation
        result = mc.hashable_representation
        self.assertEqual(hash(result), hash(ref_tuple))

    def test__mull__(self):
        # empty multicolor shall be kept as is regardless of multiplier
        mc = Multicolor()
        for multiplier in range(10):
            self.assertEqual(mc * multiplier, Multicolor())
        # multiplying by 0 shall make any multicolor an empty one
        mc1 = Multicolor(self.genome1)
        self.assertEqual(mc1 * 0, Multicolor())
        mc2 = Multicolor(self.genome1, self.genome2, self.genome3)
        self.assertEqual(mc2 * 0, Multicolor())
        mc3 = Multicolor(self.genome1, self.genome2, self.genome1)
        self.assertEqual(mc3 * 0, Multicolor())
        # multiplying by an integer shall multiply each color multiplicity respectively
        mc = Multicolor(self.genome1, self.genome2, self.genome3, self.genome1, self.genome2, self.genome1)
        for multiplier in range(1, 50):
            ref_multicolor = Multicolor()
            for _ in range(multiplier):
                ref_multicolor += mc
            self.assertEqual(mc * multiplier, ref_multicolor)

    def test__mull__incorrect(self):
        # multiplication only by integer value is allowed
        mc = Multicolor()
        for incorrect_multiplier in [.1, (1,), [1], "1"]:
            with self.assertRaises(TypeError):
                mc = mc * incorrect_multiplier

    def test_intersect(self):
        # intersection of two multicolors shall be considered in the same fashion, as intersection of sets
        # with a note, that multiplicity if shared colors shall be represented in the result
        mc1 = Multicolor(self.genome1, self.genome2, self.genome3)
        mc2 = Multicolor(self.genome1, self.genome2)
        mc3 = mc1.intersect(mc2)
        self.assertEqual(len(mc3.colors), 2)
        self.assertEqual(mc3.multicolors[self.genome1], 1)
        self.assertEqual(mc3.multicolors[self.genome2], 1)
        self.assertEqual(mc1.intersect(mc2), mc2.intersect(mc1))
        ###############################################################################################
        mc1 = Multicolor(self.genome1, self.genome1, self.genome2, self.genome3)
        mc2 = Multicolor(self.genome2, self.genome2, self.genome1, self.genome3)
        mc3 = mc1.intersect(mc2)
        self.assertEqual(len(mc2.colors), 3)
        self.assertEqual(mc3.multicolors[self.genome1], 1)  # second multicolor has only one copy of this color
        self.assertEqual(mc3.multicolors[self.genome2], 1)  # first multicolor has only one copy of this color
        self.assertEqual(mc1.intersect(mc2), mc2.intersect(mc1))
        ###############################################################################################
        mc1 = Multicolor(self.genome1, self.genome2, self.genome2, self.genome2)
        mc2 = Multicolor(self.genome2, self.genome2, self.genome2)
        mc3 = mc1.intersect(mc2)
        self.assertEqual(len(mc3.colors), 1)
        self.assertEqual(mc3.multicolors[self.genome2], 3)
        self.assertEqual(mc1.intersect(mc2), mc2.intersect(mc1))

    def test_intersect_incorrect_type(self):
        for incorrect_argument in [1, (1,), [1], "1", Mock()]:
            with self.assertRaises(TypeError):
                Multicolor().intersect(incorrect_argument)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()  # pragma: no cover
